/*
 *   empeg flash downloader (hugo@empeg.com)
 *
 *   (C) 1999 empeg ltd
 *
 *   Removed WIN32 code
 *   Removed glibc dependencies
 *   Made code style a little more consistent
 *   Added command line port selection
 *   Added reboot after flash
 *   Changed CLI options to use getopt
 *   Removed global variables
 *   There could still be a lot more error checking
 *     by Bitt Faulk <wfaulk@beaglebros.com> on April 4, 2003
 *     thanks to Jan Harkes for showing me how to deal with the serial port
 *
 *   unitdy but it works.
 */

#include <stdio.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#define MAGIC "peace..."
#define REBOOT_ATTN { 0x19, 0x0D }
#define REBOOT_ATTN_LEN (2)
#define REBOOT_MAGIC { 0x02, 0x0C, 0x01, 0x0C, 0x00, 0x2D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x6B, 0xA3 }
#define REBOOT_MAGIC_LEN (279)

/* cfmakeraw ``reimplementation'' by Bitt */ 
static int my_cfmakeraw(struct termios *termios_p)
{
	termios_p->c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
	termios_p->c_oflag &= ~(OPOST|ONLCR);
	termios_p->c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
	termios_p->c_cflag &= ~(CSIZE|PARENB);
	termios_p->c_cflag |= (CS8|CLOCAL);
	termios_p->c_cc[VMIN] = 1;
	termios_p->c_cc[VTIME] = 0;
}

static int openport(int *port, char *device)
{
	struct termios t;
	if ((*port=open(device,O_RDWR|O_NOCTTY|O_NDELAY))<0)
		return -errno;

	tcgetattr(*port,&t);
	my_cfmakeraw(&t);
	cfsetispeed(&t,B115200);
	cfsetospeed(&t,B115200);
	tcsetattr(*port,TCSANOW,&t);

	return 0;
}

static void closeport(int port)
{
	close(port);
}

static int readbyte(int port, int wait)
{
	struct timeval timeout;
	fd_set fds;

	/* Make the set */
	FD_ZERO(&fds);
	FD_SET(port,&fds);

	/* Make the timeout */
	timeout.tv_sec=wait;
	timeout.tv_usec=0;

	/* Do a select on the player input */
	if (select(port+1,&fds,NULL,NULL,&timeout))
	{
		/* Read byte */
		char d;
		read(port,&d,1);
		return(d);
	}

	/* Didn't get anything */
	return(-1);
}

static void writebyte(int port, int byte)
{
	/* Write byte to port */
	char d=byte;
	write(port,&d,1);
}

static void flushrx(int port)
{
	while(readbyte(port,0)>=0);
}

static int readhex(int port, int timeout,int digits)
{
	char buffer[16];
	int a,d,result;

	// Read hex data
	for(a=0;a<digits;a++)
	{
		if ((d=readbyte(port,timeout))<0) return(-1);
		buffer[a]=d;
	}
	buffer[digits]=0;

	// Parse it
	if (sscanf(buffer,"%x",&result)!=1) return(-2);
	return(result);
}

static int erasepage(int port, int address)
{
	int res;

	// Empty buffer first
	flushrx(port);

	// Send command
	writebyte(port, 'e');
	if (readbyte(port,1)!='e')
	{
		fprintf(stderr, "Failed to get echo-back on erase command.\n");
		return(1);
	}
  
	// Send address
	writebyte(port,(address>> 0)&0xff);
	writebyte(port,(address>> 8)&0xff);
	writebyte(port,(address>>16)&0xff);
	writebyte(port,(address>>24)&0xff);

	// Get 8-byte hex address reply and see if it matches
	if (readhex(port,1,8)!=address) return(2);

  // Should now have a '-' showing it's working on the erase
  if (readbyte(port,1)!='-') return(3);

	// Read result (leave 2 seconds)
	if ((res=readhex(port,2,2))!=0x80) return(0x100+res);

	return(0);
}

static int lockpage(int port, int address, int locktype)
{
	int res,command=(locktype==0?'u':'l');
	struct termios options;

	// Empty buffer first
	flushrx(port);
	flushrx(port);

	// Send command
	writebyte(port,command);
	tcgetattr(port, &options);
	tcsetattr(port, TCSAFLUSH, &options);
	if (fsync(port) != 0)
		printf("fsync failed\n");
	if (readbyte(port,1)!=command)
	{
		fprintf(stderr, "Failed to get echo-back on %s command.\n", (locktype==0?"unlock":"lock"));
		return(1);
	}
  
	// Send address
	writebyte(port,(address>> 0)&0xff);
	writebyte(port,(address>> 8)&0xff);
	writebyte(port,(address>>16)&0xff);
	writebyte(port,(address>>24)&0xff);

	// Get 8-byte hex address reply and see if it matches
	if (readhex(port,1,8)!=address) return(2);

	// Should now have a '-' showing it's working on the lock
	if (readbyte(port,1)!='-') return(3);

	// Read result (leave 2 seconds)
	if ((res=readhex(port,2,2))!=0x80) return(0x100+res);

	return(0);
}

static int program(int port, char *buffer,int address,int length)
{
	int checksum,a;

	// Empty buffer first
	flushrx(port);

	// Send command
	writebyte(port,'p');
	if (readbyte(port,1)!='p') return(1);

	// Send address
	writebyte(port,(address>> 0)&0xff);
	writebyte(port,(address>> 8)&0xff);
	writebyte(port,(address>>16)&0xff);
	writebyte(port,(address>>24)&0xff);

	// Get 8-char hex address reply and see if it matches
	if (readhex(port,1,8)!=address) return(2);

	// Should now have a '-' showing it's waiting for length
	if (readbyte(port,1)!='-') return(3);

	// Send length
	writebyte(port,(length>> 0)&0xff);
	writebyte(port,(length>> 8)&0xff);

	// Get 4-char hex length reply and see if it matches
	if (readhex(port,1,4)!=length) return(4);

	// Should now have a '-' showing it's waiting for data
	if (readbyte(port,1)!='-') return(5);

	// Send data
	for(a=checksum=0;a<length;a++)
	{
		checksum+=buffer[a];
		writebyte(port,buffer[a]);
	}

	// Send checksum
	writebyte(port,checksum&0xff);

	// Should get an 'ok' indicating good download
	if (readbyte(port,1)!='o') return(6);
	if (readbyte(port,1)!='k') return(7);
	if (readbyte(port,1)!='\r') return(8);
	if (readbyte(port,1)!='\n') return(9);

	// Should have program reply within a second
	if (readbyte(port,1)!='o') return(10);
	if (readbyte(port,1)!='k') return(11);

	// All ok
	return(0);
}

void usage(char *progname, int exitnum)
{
	fprintf(stderr, "Usage: %s -p <serial device> -a <address in hex> [-d] filename\n", progname);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -d: assume already in program mode\n");
	if (exitnum >= 0)
		exit(exitnum);
}

int main(int argc, char *argv[])
{
	FILE *i;
	int port;
	char *magic;
	int d,err,start,length,done,magicpos;
	int manufacturer,product;
	int erase_pos,erase_end,pagesize,lock_pos,lock_end;
	char *device;
	unsigned char reboot_attn[] = REBOOT_ATTN;
	unsigned char reboot_magic[] = REBOOT_MAGIC;
	extern char *optarg;
	extern int optind;
	int c;
	int direct=0;
	int aflag=0;
	int pflag=0;
	int errflag=0;
	struct timeval timeout;
	int j;

	while ((c = getopt(argc, argv, "p:a:d")) != EOF)
	{
		switch(c)
		{
			case 'a':
				if (aflag)
					errflag++;
				if (sscanf(optarg,"%x",&start)!=1)
				{
					fprintf(stderr,"Can't parse start address '%s'\n",optarg);
					exit(4);
				}
				aflag++;
				break;
			case 'd':
				direct++;
				break;
			case 'p':
				if (pflag)
					errflag++;
				device = optarg;
				pflag++;
				break;
			default:
				errflag++;
				break;
		}
	}

	if (errflag || !aflag || !pflag)
		usage(argv[0], 1);

	/* Open player device & setup */
	if (openport(&port, device))
	{
		fprintf(stderr,"Couldn't open device port (%s)\n",device);
		exit(2);
	}

	/* Size binary file */
	if ((i=fopen(argv[optind],"rb"))==NULL)
	{
		fprintf(stderr,"Can't open input '%s'\n",argv[optind]);
		exit(3);
	}

	if (!direct)
	{
		/* Tell user to power up the device */
		printf("Reboot empeg if it does not reboot on its own in the next few seconds\n");

		for (j=0; j<REBOOT_ATTN_LEN; j++)
		{
			writebyte(port,reboot_attn[j]);
		}

		timeout.tv_sec=0;
		timeout.tv_usec=100000;
		select(0, NULL, NULL, NULL, &timeout);

		for (j=0; j<REBOOT_MAGIC_LEN; j++)
		{
			writebyte(port,reboot_magic[j]);
		}

		magic = strdup(MAGIC);
		magicpos=0;
		while(1)
		{
			/* Look for magic string */
			d=readbyte(port,1);
			if (d<0)
			{
				magicpos=0;
			}
			else
			{
				if (d==magic[magicpos]) {
					magicpos++;
					if (magic[magicpos]==0) break;
				}
				else
				{
					magicpos=(d==magic[0])?1:0;
				}
			}
		}
		free(magic);
		
		printf("found empeg unit: entering program mode\n");
		writebyte(port,1);

		/* Manufacturer ID: Look for '#' then 4 hex digits */
		while((d=readbyte(port,1))>0 && d!='#');
		if (d<0) {
			printf("Couldn't find manufacturer ID!\n");
			exit(10);
		}
		manufacturer=readhex(port,1,4);

		/* Product ID: Look for '#' then 4 hex digits */
		while((d=readbyte(port,1))>0 && d!='#');
		if (d<0) {
			printf("Couldn't find product ID!\n");
			exit(11);
		}
		product=readhex(port,1,4);
		
		printf("manufacturer=%04x, product=%04x\n",manufacturer,product);
	}
	else /* direct */
	{
		/* Send a cr to get a prompt back */
		writebyte(port,13);
	}

	printf("waiting for prompt\n");
	while((d=readbyte(port,1))>0 && d!='?');
	if (d<0)
	{
		printf("No prompt within timeout\n");
		exit(12);
	}
	printf("starting erase [");
	fflush(stdout);

	fseek(i,0L,SEEK_END);
	length=ftell(i);
	fseek(i,0L,SEEK_SET);

	/* Erase pages that need erasing */
	lock_pos=erase_pos=(start&~0x1fff);
	lock_end=erase_end=(start+length-1)&~0x1fff;

	done=0;
	while(erase_pos<=erase_end)
	{
		/* Work out the pagesize for this bit */
		pagesize=(erase_pos<0x10000)?0x2000:0x10000;

		/* If we've got a C3 flash, we need to unlock the page first */
		if (product==0x88c1)
		{
			if ((err=lockpage(port,erase_pos,0))!=0)
			{
				fprintf(stderr,"lockpage(%x,0) got code %x\n",erase_pos,err);
				exit(5);
			}
		}

		/* Do the erase */
		if ((err=erasepage(port,erase_pos))!=0)
		{
			fprintf(stderr,"erasepage(%x) got code %x\n",erase_pos,err);
			exit(5);
		}

		erase_pos+=pagesize;
		done+=pagesize;

		printf("%3d%%\010\010\010\010",(done*100)/length);
		fflush(stdout);
	}
		 
	/* Erase done */
	printf("100%%] erase ok\nstarting program at 0x%x [",start);
	fflush(stdout);

	/* Do programming */
	done=0;
	while(done<length)
	{
		char block[16384];
		int blocklength=((length-done)>16384)?16384:(length-done);
		
		if (blocklength&1)
			blocklength++;

		/* Read block */
		fread(block,1,blocklength,i);

		/* Program data */
		if ((err=program(port,block,start,blocklength))!=0)
		{
			fprintf(stderr,"program(%x) got code %x\n",start,err);
			exit(6);
		}
		
		start+=blocklength;
		done+=blocklength;
		printf("%3d%%\010\010\010\010",(done*100)/length);
		fflush(stdout);
	}
	fclose(i);

	/* C3 flash? */
	if (product==0x88c1)
	{
		/* Lock all the pages again */
		while(lock_pos<=lock_end)
		{
			/* Work out the pagesize for this bit */
			pagesize=(lock_pos<0x10000)?0x2000:0x10000;

			if ((err=lockpage(port,lock_pos,1))!=0)
			{
				fprintf(stderr,"lockpage(%x,1) got code %x\n",lock_pos,err);
				exit(5);
			}
			lock_pos+=pagesize;
		}
	}
		
	printf("100%%] program ok\n");

	/* reboot */
	writebyte(port,'r');
	writebyte(port,'\n');

	closeport(port);
	return 0;
}

/* vim:set nowrap ts=2 sw=2: */
